using System;
using System.Collections.Generic;

namespace NumUtils.Integration
{

    public class Romberg
    {

        public double Error;
        public int N_1, N_2;
        public double FracTol;
        private double Tiny = 1e-15;

        public Romberg()
        {
            this.N_1 = 5;
            this.N_2 = 18;
            this.FracTol = 1e-6;
        }
        public Romberg(int N_1, double FractionalTolerance)
        {
            this.N_1 = N_1;
            this.N_2 = 18;
            this.FracTol = FractionalTolerance;
        }

        public double RombergClosed(double x_1, double x_2, NumUtils.DelegateFunction.DoubleFun Integrand)
        {
            if (x_1 == x_2)
            {
                return 0.0;
            }
            else
            {
                NumUtils.Interpolation.Neville INTER = new NumUtils.Interpolation.Neville();
                List<double> Integral = new List<double>();
                List<double> Delta = new List<double>();
                List<double> Answer = new List<double>();
                double Interval = x_2 - x_1;

                double h;
                double PartialArea = 0;
                double TrapeziumAreaClosed = 0;
                double M = 0;

                for (int i = 0; i < N_2; i++)
                {
                    // Begin Trapezium 
                    if (i == 0)
                    {
                        TrapeziumAreaClosed = Interval * (Integrand(x_1) + Integrand(x_2)) / 2.0; // the average of the function times the interval.
                        Delta.Add(Interval);
                    }
                    else
                    {
                        h = Interval / Math.Pow(2, i);
                        M = (int)(Math.Pow(2, i) / 2);
                        PartialArea = 0;
                        for (int k = 1; k <= M; k++)
                        {
                            PartialArea = PartialArea + Integrand(x_1 + (2 * k - 1) * h);
                        }
                        TrapeziumAreaClosed = TrapeziumAreaClosed / 2.0 + h * PartialArea;
                        Delta.Add(Interval / Math.Pow(4, i + 1));
                    }
                    // End Trapezium 
                    Integral.Add(TrapeziumAreaClosed);
                    double xx = (double)Integral[0];
                    if (i >= N_1)
                    {
                        Answer.Add(INTER.InterpolationNeville((double[])Delta.GetRange(i - N_1, N_1).ToArray(), (double[])Integral.GetRange(i - N_1, N_1).ToArray(), 0.0, N_1));
                        if (Math.Abs(INTER.ErrorInterpolationNeville / (Answer[i - N_1] + this.Tiny)) <= (FracTol + this.Tiny))
                        {
                            this.Error = INTER.ErrorInterpolationNeville;
                            return Answer[i - N_1];
                        }
                    }
                }
                throw new ApplicationException("Number of iterations exceeded in RombergClosed.");
            }
        }



        public double RombergOpenInfinity(double x_1, double x_2, NumUtils.DelegateFunction.DoubleFun Integrand)
        {
            if (x_1 == x_2)
            {
                return 0.0;
            }
            else
            {
                NumUtils.Interpolation.Neville INTER = new NumUtils.Interpolation.Neville();
                List<double> Integral = new List<double>();
                List<double> Delta = new List<double>();
                List<double> Answer = new List<double>();

                double h = 0;
                double X1 = 1 / x_2;
                double X2 = 1 / x_1;
                double Interval = X2 - X1;
                double PartialArea = 0;
                double TrapeziumAreaOpenInfinite = 0;
                double M = 0;
                for (int i = 0; i < N_2; i++)
                {
                    // Begin Trapezium Integration
                    if (i == 0)
                    {
                        TrapeziumAreaOpenInfinite = Interval * (IntegrandInverseFunction(Integrand, (X1 + X2) / 2.0));
                        Delta.Add(Interval);
                    }
                    else
                    {
                        M = (int)(Math.Pow(3, i - 1));
                        h = Interval / (M * 3);
                        PartialArea = 0;
                        for (int k = 1; k <= M; k++)
                        {
                            PartialArea = PartialArea + IntegrandInverseFunction(Integrand, X1 + 0.5 * h * (1 + (k - 1) * 6));
                            PartialArea = PartialArea + IntegrandInverseFunction(Integrand, X1 + 0.5 * h * (5 + (k - 1) * 6));
                        }
                        TrapeziumAreaOpenInfinite = TrapeziumAreaOpenInfinite / 3.0 + h * PartialArea;
                        Delta.Add(Interval / Math.Pow(9, i + 1));
                    }
                    // End Trapezium Integration
                    Integral.Add(TrapeziumAreaOpenInfinite);
                    if (i >= N_1)
                    {
                        Answer.Add(INTER.InterpolationNeville((double[])Delta.GetRange(i - N_1, N_1).ToArray(), (double[])Integral.GetRange(i - N_1, N_1).ToArray(), 0.0, N_1));
                        if (Math.Abs(INTER.ErrorInterpolationNeville / (Answer[i - N_1] + this.Tiny)) <= (FracTol + this.Tiny))
                        {
                            this.Error = INTER.ErrorInterpolationNeville;
                            return Answer[i - N_1];
                        }
                    }
                }
                throw new ApplicationException("Number of iterations exceeded in RombergOpenInfinity.");
            }
        }

        public double IntegrandInverseFunction(NumUtils.DelegateFunction.DoubleFun INTEGRAND, double x)
        {
            return INTEGRAND(1 / x) / x / x;
        }




    }
}

