using System;
using System.Collections;

namespace NumUtils.RootFinding
{

    public class BracketingWithNewtonRaphson
    {
        public double Error;
        public double ExpansionFactor;
        public int NumVertices;
        public double TOL;
        public double MAXIT;
        private double Tiny = 1e-15;

        public BracketingWithNewtonRaphson()
        {
            this.ExpansionFactor = 5.0;
            this.NumVertices = 5;
            this.TOL = 1e-10;
            this.MAXIT = 4000;
        }

        public BracketingWithNewtonRaphson(double TOL)
        {
            this.ExpansionFactor = 10.0;
            this.NumVertices = 5;
            this.TOL = TOL;
            this.MAXIT = 4000;
        }

        public BracketingWithNewtonRaphson(double ExpansionFactor, int NumVertices, double TOL, int MAXIT)
        {
            this.ExpansionFactor = ExpansionFactor;
            this.NumVertices = NumVertices;
            this.TOL = TOL;
            this.MAXIT = MAXIT;
        }



        public double NewtonRaphson(NumUtils.DelegateFunction.DoubleFun FUNCTION, NumUtils.DelegateFunction.DoubleFun FUNCTION_Derivative, double x)
        {
            return x - FUNCTION(x) / FUNCTION_Derivative(x);
        }

        public double NewtonRaphson(NumUtils.DelegateFunction.DoubleFun FUNCTION, NumUtils.DelegateFunction.DoubleFun FUNCTION_Derivative, double x, double y)
        {
            return x - (FUNCTION(x)-y) / FUNCTION_Derivative(x);
        }

        public double NewtonRaphson(double FUNCTION, double FUNCTION_Derivative, double x, double y)
        {
            return x - (FUNCTION - y) / FUNCTION_Derivative;
        }

        public int[] GetBoundingIndexes(double[] X, double[] Y, double y, int n)
        {
            ArrayList X1 = new ArrayList(X);
            ArrayList X2 = new ArrayList(X);
            int[] Ans = new int[2];
            X1 = X1.GetRange(0, n);
            X2 = X2.GetRange(0, n);
            X1.Sort();
            for (int i = 0; i < X1.Count - 1; i++)
            {
                if (Y[X2.IndexOf(X1[i])] <= y && y <= Y[X2.IndexOf(X1[i + 1])])
                {
                    return new int[2] { X2.IndexOf(X1[i]), X2.IndexOf(X1[i + 1]) };
                }
                if (Y[X2.IndexOf(X1[i])] >= y && y >= Y[X2.IndexOf(X1[i + 1])])
                {
                    return new int[2] { X2.IndexOf(X1[i]), X2.IndexOf(X1[i + 1]) };
                }
            }
            throw new ApplicationException("Problem in GetBoundingIndexes.");
        }

        public double GetRootWithNewtonRaphsonAndNeville(NumUtils.DelegateFunction.DoubleFun FUNCTION, NumUtils.DelegateFunction.DoubleFun FUNCTION_DERIVATIVE, double y, double X0, double X1)
        {
            double[] X = new double[2] { X0, X1 }; ;// new double[2] { 1, 1 };// initial guess in redshift
            double[] Y = new double[2];
            double StoredX=0;
            double StoredY=0;
            double[] PointsY = new double[NumVertices + 3];
            double[] PointsX = new double[NumVertices + 3];

            double DeltaX;
            double X_0, Y_0;
            double X_1, Y_1;

            double XX=0;
            double YY = 0; 
            double StoredY_P = 0;

            Y[0] = FUNCTION(X[0]);
            Y[1] = FUNCTION(X[1]);

            int flag = 0;
            NumUtils.Interpolation.Neville Neville = new NumUtils.Interpolation.Neville();
            for (int ScapeIndex = 0; ScapeIndex < this.MAXIT; ScapeIndex++)
            {

                if ((y >= Y[0] && y <= Y[1]) || (y >= Y[1] && y <= Y[0]))
                {
                    if (Math.Abs(2 * (X[1] - X[0]) / (X[1] + X[0] + Tiny)) < TOL + Tiny) { this.Error = X[1] - X[0]; return (X[0] + X[1]) / 2; }

                    if (flag == 0) 
                    { 
                        StoredX = (X[0] + X[1]) / 2; 
                        StoredY = FUNCTION(StoredX);
                        StoredY_P = FUNCTION_DERIVATIVE(StoredX);
                        flag = 1;
                    }

                    XX = StoredX; YY = StoredY;
                    StoredX = NewtonRaphson(StoredY, StoredY_P, StoredX, y);

                    double del = Math.Abs(2 * StoredY / StoredY_P);
                    if ((StoredX >= X[0] && StoredX <= X[1]) )
                    {
                        if (Math.Abs(2 * (XX - StoredX) / (XX + StoredX + Tiny)) < (TOL + Tiny))
                        {
                            this.Error = Math.Abs(StoredX - XX);
                            return StoredX;
                        }
                        StoredY = FUNCTION(StoredX);
                        StoredY_P = FUNCTION_DERIVATIVE(StoredX);
                        if (Y[0] < Y[1])
                        {
                            if (y >= StoredY) { X[0] = StoredX; Y[0] = StoredY; }
                            else { X[1] = StoredX; Y[1] = StoredY; }
                        }
                        else
                        {
                            if (y >= StoredY) { X[1] = StoredX; Y[1] = StoredY; }
                            else { X[0] = StoredX; Y[0] = StoredY; }
                        }
                    }
                    else
                    {
                        flag = 0;
                        DeltaX = (X[1] - X[0]) / (NumVertices - 1);
                        PointsX[0] = X[0];
                        PointsY[0] = Y[0];
                        PointsX[NumVertices - 1] = X[1];
                        PointsY[NumVertices - 1] = Y[1];
                        if (Math.Abs(2 * (y - PointsY[0]) / (y + PointsY[0] + Tiny)) < TOL + Tiny) { this.Error = PointsX[0] * (TOL + Tiny); return PointsX[0]; }
                        if (Math.Abs(2 * (y - PointsY[NumVertices - 1]) / (y + PointsY[NumVertices - 1] + Tiny)) < TOL + Tiny) { this.Error = PointsX[NumVertices - 1] * (TOL + Tiny); return PointsX[NumVertices - 1]; }

                        for (int i = 1; i < (NumVertices - 1); i++)
                        {
                            if (i == (NumVertices - 1) / 2) { PointsX[i] = XX; PointsY[i] = YY; }
                            else { PointsX[i] = X[0] + i * DeltaX; PointsY[i] = FUNCTION(PointsX[i]); }
                            if (Math.Abs(2 * (y - PointsY[i]) / (y + PointsY[i] + Tiny)) < TOL + Tiny) { this.Error = PointsX[i] * (TOL + Tiny); return PointsX[i]; }
                        }

                        StoredX = Neville.InterpolationNeville(PointsY, PointsX, y, NumVertices);

                        X_0 = StoredX - Neville.ErrorInterpolationNeville;
                        X_1 = StoredX + Neville.ErrorInterpolationNeville;
                        if ((X[0] <= X_0 && X_1 <= X[1]))
                        {
                            if (Math.Abs(Neville.ErrorInterpolationNeville / (StoredX + Tiny)) < (TOL + Tiny))
                            {
                                this.Error = Math.Abs(Neville.ErrorInterpolationNeville);
                                return StoredX;
                            }

                            Y_0 = FUNCTION(X_0);
                            Y_1 = FUNCTION(X_1);
                            StoredY = FUNCTION(StoredX);
                            if ((Y_0 <= y && y <= Y_1) || (Y_1 <= y && y <= Y_0))
                            {
                                if (Y_0 < Y_1)
                                {
                                    if (y >= StoredY) { X[0] = StoredX; X[1] = X_1; Y[0] = StoredY; Y[1] = Y_1; }
                                    else { X[0] = X_0; X[1] = StoredX; Y[0] = Y_0; Y[1] = StoredY; }
                                }
                                else
                                {
                                    if (y >= StoredY) { X[0] = X_0; X[1] = StoredX; Y[0] = Y_1; Y[1] = StoredY; }
                                    else { X[0] = StoredX; X[1] = X_1; Y[0] = StoredY; Y[1] = Y_1; }
                                }
                            }
                            else
                            {
                                PointsX[NumVertices] = X_0; PointsY[NumVertices] = Y_0;
                                PointsX[NumVertices + 1] = X_1; PointsY[NumVertices + 1] = Y_1;
                                PointsX[NumVertices + 2] = StoredX; PointsY[NumVertices + 2] = StoredY;
                                int[] BoundingIndexes = GetBoundingIndexes(PointsX, PointsY, y, NumVertices + 3);
                                X[0] = PointsX[BoundingIndexes[0]];
                                X[1] = PointsX[BoundingIndexes[1]];
                                Y[0] = PointsY[BoundingIndexes[0]];
                                Y[1] = PointsY[BoundingIndexes[1]];
                            }
                        }
                        else
                        {
                            int[] BoundingIndexes = GetBoundingIndexes(PointsX, PointsY, y, NumVertices);
                            X[0] = PointsX[BoundingIndexes[0]];
                            X[1] = PointsX[BoundingIndexes[1]];
                            Y[0] = PointsY[BoundingIndexes[0]];
                            Y[1] = PointsY[BoundingIndexes[1]];
                        }
                    }
                }//starts the bracketing
                else
                {
                    if (Y[1] > Y[0])
                    {
                        if (y > Y[1])
                        {
                            X[0] = X[1]; Y[0] = Y[1];
                            X[1] = X[1] * this.ExpansionFactor; Y[1] = Y[1] = FUNCTION(X[1]);
                        }
                        else
                        {
                            X[1] = X[0]; Y[1] = Y[0];
                            X[0] = X[0] / ExpansionFactor; Y[0] = FUNCTION(X[0]);
                        }
                    }
                    else
                    {
                        if (y > Y[0])
                        {
                            X[1] = X[0]; Y[1] = Y[0];
                            X[0] = X[0] / this.ExpansionFactor; Y[0] = FUNCTION(X[0]);
                        }
                        else
                        {
                            X[0] = X[1]; Y[0] = Y[1];
                            X[1] = X[1] * ExpansionFactor; Y[1] = FUNCTION(X[1]);
                        }
                    }
                }

            }
            throw new ApplicationException("Number of iterations exceeded in GetRootWithNewtonRaphsonAndNeville.");
        }




    
    
    }
}
