using System.Collections;
using System;
using System.Data;
using System.Data.SqlClient;
using System.Data.SqlTypes;
using Microsoft.SqlServer.Server;

namespace SQLCosmology
{


    public class SQLCosmologicalFunctionsFreeParameters
    {

        /*
                See this paper for a better review of the parameters and equations:
                "FIVE-YEAR WILKINSON MICROWAVE ANISOTROPY PROBE OBSERVATIONS:COSMOLOGICAL INTERPRETATION" Komatsu et al. (2009) ApJS 180:330-376
		        that can be downloaded from:
		        http://lambda.gsfc.nasa.gov/product/map/dr3/map_bibliography.cfm
        */

        public const double TOL_Romberg = 1e-9;
        public const int N_1_Romberg = 5;
        public const double TOL_BracketingWithNeville = 1e-9;
        public const double ExpansionFactor_BracketingWithNeville = 10.0;// to be used in the BracketingWithNeville routine.
        public const int NumVertices_BracketingWithNeville = 5;
        public const int MAXIT_BracketingWithNeville = 5000;


        [SqlFunction(
        Name = "fCosmoZfromTh",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble RedshiftAtHubbleTime(SqlDouble HubbleTIME, double OmegaM, double OmegaL, double OmegaR, double w_0, double h_0)
        {
            SqlDouble answer;
            try
            {
                if (HubbleTIME.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.RedshiftAtHubbleTime((double)HubbleTIME);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmoZfromDh",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble RedshiftAtHubbleDistance(SqlDouble HubbleDISTANCE, double OmegaM, double OmegaL, double OmegaR, double w_0, double h_0)
        {
            SqlDouble answer;
            try
            {
                if (HubbleDISTANCE.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.RedshiftAtHubbleDistance((double)HubbleDISTANCE);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }




        [SqlFunction(
        Name = "fCosmoZfromDl",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble RedshiftAtLuminosityDistance(SqlDouble LuminosityDistance, double OmegaM, double OmegaL, double OmegaR, double w_0, double h_0)
        {
            SqlDouble answer;
            try
            {
                if (LuminosityDistance.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.RedshiftAtLuminosityDistance((double)LuminosityDistance);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmoComovVolumeFromDl",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble ComovVolumeAtLumDist(SqlDouble LuminosityDistance, double OmegaM, double OmegaL, double OmegaR, double w_0, double h_0)
        {
            SqlDouble answer;
            try
            {
                if (LuminosityDistance.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.ComovVolumeAtLumDist((double)LuminosityDistance);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmoZfromComovVolume",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble RedshiftAtComovVolume(SqlDouble ComovVolume, double OmegaM, double OmegaL, double OmegaR, double w_0, double h_0)
        {
            SqlDouble answer;
            try
            {
                if (ComovVolume.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.RedshiftAtComovVolume((double)ComovVolume);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        public static void FillRowRedshiftAtAngularDiamDist(object item, out SqlDouble z1, out SqlDouble z2)
        {
            double[] row = (double[])item;
            z1 = (SqlDouble)row[0];
            z2 = (SqlDouble)row[1];
        }

        [SqlFunction(
        Name = "fCosmoZfromDa",
        FillRowMethodName = "FillRowRedshiftAtAngularDiamDist",
        TableDefinition = "z1 float, z2 float",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static IEnumerable RedshiftAtAngularDiamDist(SqlDouble AngularDiamDist, double OmegaM, double OmegaL, double OmegaR, double w_0, double h_0)
        {
            ArrayList RedshiftArray = new ArrayList(1);
            try
            {
                if (AngularDiamDist.IsNull)
                    RedshiftArray.Add(new SqlDouble[] { SqlDouble.Null, SqlDouble.Null });
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    RedshiftArray.Add(Cosmo.RedshiftAtAngularDiamDist((double)AngularDiamDist));
                }
            }
            catch
            {
                throw;
            }
            return RedshiftArray;
        }

        [SqlFunction(
        Name = "fCosmoZfromDm",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble RedshiftAtComovDistTransverse(SqlDouble ComovDistTransverse, double OmegaM, double OmegaL, double OmegaR, double w_0, double h_0)
        {
            SqlDouble answer;
            try
            {
                if (ComovDistTransverse.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.RedshiftAtComovDistTransverse((double)ComovDistTransverse);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmoZfromDc",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble RedshiftAtComovDistLineOfSight(SqlDouble ComovDistLineOfSight, double OmegaM, double OmegaL, double OmegaR, double w_0, double h_0)
        {
            SqlDouble answer;
            try
            {
                if (ComovDistLineOfSight.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.RedshiftAtComovDistLineOfSight((double)ComovDistLineOfSight);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }


        [SqlFunction(
        Name = "fCosmZfromAgeOfUniverse",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false)]
        public static SqlDouble RedshiftAtAgeOfUniverse(SqlDouble AgeOfUniverse, double OmegaM, double OmegaL, double OmegaR, double w_0, double h_0)
        {
            SqlDouble answer;
            try
            {
                if (AgeOfUniverse.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.RedshiftAtAgeOfUniverse((double)AgeOfUniverse);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmZfromLookBackTime",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble RedshiftAtLookBackTime(SqlDouble LookBackTime, double OmegaM, double OmegaL, double OmegaR, double w_0, double h_0)
        {
            SqlDouble answer;
            try
            {
                if (LookBackTime.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.RedshiftAtLookBackTime((double)LookBackTime);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmoLookBackTime",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble LookBackTime(SqlDouble z, double OmegaM, double OmegaL, double OmegaR, double w_0, double h_0)
        {
            SqlDouble answer;
            try
            {
                if (z.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.LookBackTime((double)z);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }



        [SqlFunction(
        Name = "fCosmoTimeInterval",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble TimeInterval(SqlDouble zMin, SqlDouble zMax, double OmegaM, double OmegaL, double OmegaR, double w_0, double h_0)
        {
            SqlDouble answer;
            try
            {
                if (zMin.IsNull || zMax.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.TimeInterval((double)zMin, (double)zMax);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }



        [SqlFunction(
        Name = "fCosmoAgeOfUniverse",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble AgeOfUniverse(SqlDouble z, double OmegaM, double OmegaL, double OmegaR, double w_0, double h_0)
        {
            SqlDouble answer;
            try
            {
                if (z.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.AgeOfUniverse((double)z);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }


        [SqlFunction(
        Name = "fCosmoHubbleDistance",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble HubbleDistance(SqlDouble z, double OmegaM, double OmegaL, double OmegaR, double w_0, double h_0)
        {
            SqlDouble answer;
            try
            {
                if (z.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.HubbleDistance((double)z);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmoHubbleTime",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble HubbleTime(SqlDouble z, double OmegaM, double OmegaL, double OmegaR, double w_0, double h_0)
        {
            SqlDouble answer;
            try
            {
                if (z.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.HubbleTime((double)z);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }
        
        
        [SqlFunction(
        Name = "fCosmoDl",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble LuminosityDistance(SqlDouble z, double OmegaM, double OmegaL, double OmegaR, double w_0, double h_0)
        {
            SqlDouble answer;
            try
            {
                if (z.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.LuminosityDistance((double)z);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmoDc",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble ComovDistLineOfSight(SqlDouble z, double OmegaM, double OmegaL, double OmegaR, double w_0, double h_0)
        {
            SqlDouble answer;
            try
            {
                if (z.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.ComovDistLineOfSight(0, (double)z);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmoDcInterval",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble ComovDistLineOfSightInterval(SqlDouble Zmin, SqlDouble Zmax, double OmegaM, double OmegaL, double OmegaR, double w_0, double h_0)
        {
            SqlDouble answer;
            try
            {
                if (Zmin.IsNull || Zmax.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.ComovDistLineOfSight((double)Zmin, (double)Zmax);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmoComovDist2ObjectsRaDec",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble ComovDistBetween2objectsRaDec(SqlDouble Redshift1, SqlDouble RA1, SqlDouble DEC1, SqlDouble Redshift2, SqlDouble RA2, SqlDouble DEC2, double OmegaM, double OmegaL, double OmegaR, double w_0, double h_0)
        {
            SqlDouble answer;
            try
            {
                if (Redshift1.IsNull || Redshift2.IsNull || RA1.IsNull || DEC1.IsNull || RA2.IsNull || DEC2.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.ComovDistBetween2objectsRaDec((double)Redshift1, (double)RA1, (double)DEC1, (double)Redshift2, (double)RA2, (double)DEC2);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmoComovDist2ObjectsXYZ",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble ComovDistBetween2objectsXyz(SqlDouble Redshift1, SqlDouble x1, SqlDouble y1, SqlDouble z1, SqlDouble Redshift2, SqlDouble x2, SqlDouble y2, SqlDouble z2, double OmegaM, double OmegaL, double OmegaR, double w_0, double h_0)
        {
            SqlDouble answer;
            try
            {
                if (Redshift1.IsNull || Redshift2.IsNull || x1.IsNull || y1.IsNull || z1.IsNull || x2.IsNull || y2.IsNull || z2.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.ComovDistBetween2objectsXyz((double)Redshift1, (double)x1, (double)y1, (double)z1, (double)Redshift2, (double)x2, (double)y2, (double)z2);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmoComovDist2Objects",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble ComovDistBetween2objects(SqlDouble Redshift1, SqlDouble Redshift2, SqlDouble SeparationAngle, double OmegaM, double OmegaL, double OmegaR, double w_0, double h_0)
        {
            SqlDouble answer;
            try
            {
                if (Redshift1.IsNull || Redshift2.IsNull || SeparationAngle.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.ComovDistBetween2objects((double)Redshift1, (double)Redshift2, (double)SeparationAngle);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fMathAngSepXYZ",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble AngularSeparationXyz(SqlDouble x1, SqlDouble y1, SqlDouble z1, SqlDouble x2, SqlDouble y2, SqlDouble z2)
        {
            SqlDouble answer;
            try
            {
                if (x1.IsNull || y1.IsNull || z1.IsNull || x2.IsNull || y2.IsNull || z2.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Mathematics.MathFunctions MATH = new Mathematics.MathFunctions();
                    answer = MATH.AngularSeparationXyz((double)x1, (double)y1, (double)z1, (double)x2, (double)y2, (double)z2);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fMathAngSepRADEC",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble AngularSeparationRaDec(SqlDouble RA_1, SqlDouble DEC_1, SqlDouble RA_2, SqlDouble DEC_2)
        {
            SqlDouble answer;
            try
            {
                if (RA_1.IsNull || DEC_1.IsNull || RA_2.IsNull || DEC_2.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Mathematics.MathFunctions MATH = new Mathematics.MathFunctions();
                    answer = MATH.AngularSeparationRaDec((double)RA_1, (double)DEC_1, (double)RA_2, (double)DEC_2);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmfDa",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble ComovDistTransverse(SqlDouble z, double OmegaM, double OmegaL, double OmegaR, double w_0, double h_0)
        {
            SqlDouble answer;
            try
            {
                if (z.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.ComovDistTransverse((double)z);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmoDm",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble AngularDiameterDist(SqlDouble z, double OmegaM, double OmegaL, double OmegaR, double w_0, double h_0)
        {
            SqlDouble answer;
            try
            {
                if (z.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.AngularDiameterDist((double)z);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmoComovingVolume",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble ComovVolume(SqlDouble z, double OmegaM, double OmegaL, double OmegaR, double w_0, double h_0)
        {
            SqlDouble answer;
            try
            {
                if (z.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.ComovVolume((double)z);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmoAbsMagFromLumDist",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble AbsMagFromLumDist(SqlDouble m, SqlDouble LuminosityDistance)
        {
            SqlDouble answer;
            try
            {
                if (LuminosityDistance.IsNull || m.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions();
                    answer = Cosmo.AbsMagFromLumDist((double)m, (double)LuminosityDistance, 0.0);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }

        [SqlFunction(
        Name = "fCosmoAbsMag",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static SqlDouble AbsMag(SqlDouble m, SqlDouble z, double OmegaM, double OmegaL, double OmegaR, double w_0, double h_0)
        {
            SqlDouble answer;
            try
            {
                if (z.IsNull || m.IsNull)
                    answer = SqlDouble.Null;
                else
                {
                    Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                    answer = Cosmo.AbsMag((double)m, (double)z, 0.0);
                }
            }
            catch
            {
                throw;
            }
            return answer;
        }


        public static void FillRowGridAllCosmoQuantitiesIEnumerable(object item, out SqlDouble z, out SqlDouble Dc, out SqlDouble Dm, out SqlDouble Da, out SqlDouble Dl, out SqlDouble ComVol, out SqlDouble LookBackTime, out SqlDouble AgeOfUniverse)
        {
            double[] row = (double[])(item);
            z = SqlDouble.Parse(row[0].ToString());
            Dc = SqlDouble.Parse(row[1].ToString());
            Dm = SqlDouble.Parse(row[2].ToString());
            Da = SqlDouble.Parse(row[3].ToString());
            Dl = SqlDouble.Parse(row[4].ToString());
            ComVol = SqlDouble.Parse(row[5].ToString());
            LookBackTime = SqlDouble.Parse(row[6].ToString());
            AgeOfUniverse = SqlDouble.Parse(row[7].ToString());
        }

        [SqlFunction(
        Name = "fCosmoAllQuantities",
        FillRowMethodName = "FillRowGridAllCosmoQuantitiesIEnumerable",
        TableDefinition = "z float, Dc float, Dm float, Da float, Dl float, ComovingVolume float, LookBackTime float, AgeOfUniverse float",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static IEnumerable GridAllCosmoQuantitiesIEnumerable(SqlDouble zMin, SqlDouble zMax, int NumCell, double OmegaM, double OmegaL, double OmegaR, double w_0, double h_0)
        {
            if (zMin.IsNull || zMax.IsNull)
                yield return new SqlDouble[] { SqlDouble.Null, SqlDouble.Null, SqlDouble.Null, SqlDouble.Null, SqlDouble.Null, SqlDouble.Null, SqlDouble.Null, SqlDouble.Null };
            else
            {
                Cosmology.CosmologicalFunctions Cosmo = new Cosmology.CosmologicalFunctions(OmegaM, OmegaL, OmegaR, w_0, h_0, TOL_Romberg, N_1_Romberg, TOL_BracketingWithNeville, ExpansionFactor_BracketingWithNeville, NumVertices_BracketingWithNeville, MAXIT_BracketingWithNeville);
                foreach (double[] row in Cosmo.GridAllCosmoQuantities((double)zMin, (double)zMax, NumCell))
                {
                    yield return row;
                }
            }
        }

        static public string Revision
        {
            get { return "$Revision: 1.0 $"; }
        }

        /* Revision History

                $Log: msdevcvs.aspx,v $
                Revision 1.1  2008/06/20 10:59:55  mtaghiza
                 Created Revision and History for this class.
        */



    }




}




