using System;
using System.Data;
using System.Data.SqlClient;
using System.Data.SqlTypes;
using Microsoft.SqlServer.Server;
using System.Collections;
using System.Collections.Generic;
using System.IO;


namespace SQLMath
{

    public partial class SQLMathFunctions
    {

        public const int MaxNumCells_Hist = (int)1e7;
        public const int MaxNumRows_Hist = (int)1e7;
        public const int MaxNumRows_Stat = (int)1e7;
        public const int MaxNumRows_Grid = (int)1e9;

        //----------------------------------------------------------------------------------------------------------------------------------------------------

        public static void FillRowGrid(object item, out SqlDouble X)
        {
            double[] row = (double[])(item);
            X = SqlDouble.Parse(row[0].ToString());
        }

        [SqlFunction(
        Name = "fMathGrid",
        FillRowMethodName = "FillRowGrid",
        TableDefinition = "TickMark float primary key not null",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None,
        IsDeterministic = true,
        IsPrecise = false
        )]
        public static IEnumerable Grid(SqlDouble X1, SqlDouble X2, int NumCell, bool IsLinear, bool IsMidPoints)
        {
            if (NumCell > MaxNumRows_Grid)
                throw new System.Exception("Invalid parameter in  SQLmath.Grid. The value of NumCell should be smaller than " + MaxNumRows_Grid.ToString().Trim());
            if (X1.IsNull || X2.IsNull)
                yield return new SqlDouble[] { SqlDouble.Null };
            else
            {
                Mathematics.MathFunctions MATH = new Mathematics.MathFunctions();
                foreach (double[] row in MATH.Grid((double)X1, (double)X2, NumCell, IsLinear, IsMidPoints))
                {
                    yield return row;
                }
            }
        }

        //----------------------------------------------------------------------------------------------------------------------------------------------------

        [SqlFunction(
        Name = "fMathRound",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None
        )]
        public static SqlDouble Round(SqlDouble x, int FractionalSignificance)
        {
            if (!x.IsNull)
            {
                Mathematics.MathFunctions math = new Mathematics.MathFunctions();
                return math.Round((double)x, FractionalSignificance);
            }
            else
                return SqlDouble.Null;
        }

        //----------------------------------------------------------------------------------------------------------------------------------------------------

        [SqlFunction(
        Name = "fMathBin",
        DataAccess = DataAccessKind.None,
        SystemDataAccess = SystemDataAccessKind.None
        )]
        public static SqlDouble Bin(SqlDouble X, SqlDouble X1, SqlDouble X2, int NumCell, bool HasOpenUpperBound, bool IsLinearScale)
        {
            if (X.IsNull || X1.IsNull || X2.IsNull)
                return SqlDouble.Null;
            else
            {
                Mathematics.MathFunctions math = new Mathematics.MathFunctions();
                return math.Bin((double)X, (double)X1, (double)X2, NumCell, HasOpenUpperBound, IsLinearScale);
            }
        }

        //----------------------------------------------------------------------------------------------------------------------------------------------------

        [SqlProcedure(Name = "spMathHistogramNDim")]
        public static void HistogramNDimFromSafeQuery(string Query, int Dimension, string String_X1, string String_X2, string String_NumCell, bool HasOpenUpperBound)
        {
            string Query2 = "select dbo.fMathSafeSqlCommand('" + Query + "'," + MaxNumRows_Hist.ToString().Trim() + ")";
            string SafeQuery;
            using (SqlConnection connection = new SqlConnection("context connection=true"))
            using (SqlCommand cmd = new SqlCommand(Query2, connection))
            {
                connection.Open();
                SqlDataReader rdr = cmd.ExecuteReader();
                rdr.Read();
                string ColName = rdr.GetName(0);
                SafeQuery = rdr[0].ToString();
                if (SafeQuery.StartsWith("error") || ColName == "error_message" || SafeQuery.StartsWith("Syntax is OK"))
                    throw new System.Exception(SafeQuery);
                rdr.Close();
                connection.Close();
            }
            HistogramNDimFromUnSafeQuery(SafeQuery, Dimension, String_X1, String_X2, String_NumCell, HasOpenUpperBound);
        }

        //----------------------------------------------------------------------------------------------------------------------------------------------------

        public static void HistogramNDimFromUnSafeQuery(string Query, int Dimension, string String_X1, string String_X2, string String_NumCell, bool HasOpenUpperBound)
        {
            if (Dimension < 1)
                throw new System.Exception("Invalid parameter in  SQLmath.SQLMathFunctions. The value of Dimension should be greater or equal than 1.");
            Mathematics.MathFunctions math = new Mathematics.MathFunctions();
            ArrayList Data = new ArrayList(Dimension);
            for (int i = 0; i < Dimension; i++)
            {
                Data.Add(new ArrayList());
            }

            ArrayList Weights = new ArrayList();
            using (SqlConnection connection = new SqlConnection("context connection=true"))
            using (SqlCommand cmd = new SqlCommand(Query, connection))
            {
                try
                {
                    double[,] Hist = math.HistogramNDim(Dimension, connection, cmd, String_X1, String_X2, String_NumCell, HasOpenUpperBound, MaxNumCells_Hist);

                    SqlMetaData[] columns = new SqlMetaData[Dimension + 2];
                    for (int i = 0; i < columns.GetLength(0) - 2; i++)
                    {
                        columns[i] = new SqlMetaData("BinCenter_" + (i + 1).ToString().Trim(), SqlDbType.Float);
                    }
                    columns[columns.GetLength(0) - 2] = new SqlMetaData("Height", SqlDbType.Float);
                    columns[columns.GetLength(0) - 1] = new SqlMetaData("Counts", SqlDbType.Float);

                    SqlDataRecord record = new SqlDataRecord(columns);
                    SqlContext.Pipe.SendResultsStart(record);

                    for (int i = 0; i < Hist.GetLength(0); i++)
                    {
                        for (int j = 0; j < Dimension + 2; j++)
                        {
                            record.SetSqlDouble(j, (SqlDouble)(Hist[i, j]));
                        }
                        SqlContext.Pipe.SendResultsRow(record);
                    }
                    SqlContext.Pipe.SendResultsEnd();
                }
                catch
                {
                    throw;
                }
                finally
                {
                    connection.Close();
                }
            }
        }

        //----------------------------------------------------------------------------------------------------------------------------------------------------



        public static void HistogramNDimFromUnSafeQuery2(string Query, int Dimension, string String_X1, string String_X2, string String_NumCell, bool HasOpenUpperBound)
        {
            if (Dimension < 1)
                throw new System.Exception("Invalid parameter in  SQLmath.SQLMathFunctions. The value of Dimension should be greater or equal than 1.");
            Mathematics.MathFunctions math = new Mathematics.MathFunctions();
            ArrayList Data = new ArrayList(Dimension);
            for (int i = 0; i < Dimension; i++)
            {
                Data.Add(new ArrayList());
            }

            ArrayList Weights = new ArrayList();
            using (SqlConnection connection = new SqlConnection("context connection=true"))
            using (SqlCommand cmd = new SqlCommand(Query, connection))
            {
                try
                {
                    connection.Open();
                    SqlDataReader rdr = cmd.ExecuteReader();
                    rdr.Read();

                    //string ColName = rdr.GetName(0);
                    string QueryResultString = rdr[0].ToString();

                    if (rdr.FieldCount > (Dimension + 1) || rdr.FieldCount == 0 || rdr.FieldCount < Dimension)
                        throw new System.Exception("Incorrect number of query result columns or Dimension value in SQLmath.SQLMathFunctions.HistogramNDimFromQuery. Number of Columns returned by QueryToTable (=" + rdr.FieldCount.ToString().Trim() + ") should be equal to either " + Dimension.ToString().Trim() + " or " + ((int)(Dimension + 1)).ToString().Trim());

                    if (rdr.FieldCount == Dimension)
                    {
                        for (int i = 0; i < Dimension; i++) { ((ArrayList)Data[i]).Add(double.Parse(rdr[i].ToString())); }
                        Weights.Add(1.0);
                        while (rdr.Read())
                        {
                            for (int i = 0; i < Dimension; i++) { ((ArrayList)Data[i]).Add(double.Parse(rdr[i].ToString())); }
                            Weights.Add(1.0);
                        }
                    }
                    else
                    {
                        for (int i = 0; i < Dimension; i++) { ((ArrayList)Data[i]).Add(double.Parse(rdr[i].ToString())); }
                        Weights.Add(double.Parse(rdr[Dimension].ToString()));
                        while (rdr.Read())
                        {
                            for (int i = 0; i < Dimension; i++) { ((ArrayList)Data[i]).Add(double.Parse(rdr[i].ToString())); }
                            Weights.Add(double.Parse(rdr[Dimension].ToString()));
                        }
                    }

                    rdr.Close();
                    double[,] Hist = math.HistogramNDim(Dimension, Data, Weights, String_X1, String_X2, String_NumCell, HasOpenUpperBound, MaxNumCells_Hist);

                    SqlMetaData[] columns = new SqlMetaData[Dimension + 2];
                    for (int i = 0; i < columns.GetLength(0) - 2; i++)
                    {
                        columns[i] = new SqlMetaData("BinCenter_" + (i + 1).ToString().Trim(), SqlDbType.Float);
                    }
                    columns[columns.GetLength(0) - 2] = new SqlMetaData("Height", SqlDbType.Float);
                    columns[columns.GetLength(0) - 1] = new SqlMetaData("Counts", SqlDbType.Float);

                    SqlDataRecord record = new SqlDataRecord(columns);
                    SqlContext.Pipe.SendResultsStart(record);

                    for (int i = 0; i < Hist.GetLength(0); i++)
                    {
                        for (int j = 0; j < Dimension + 2; j++)
                        {
                            record.SetSqlDouble(j, (SqlDouble)(Hist[i, j]));
                        }
                        SqlContext.Pipe.SendResultsRow(record);
                    }
                    SqlContext.Pipe.SendResultsEnd();
                }
                catch
                {
                    throw;
                }
                finally
                {
                    connection.Close();
                }
            }
        }

        //----------------------------------------------------------------------------------------------------------------------------------------------------



        [SqlFunction(
        Name = "fMathQuantile",
        DataAccess = DataAccessKind.Read,
        SystemDataAccess = SystemDataAccessKind.Read,
        IsDeterministic = false,
        IsPrecise = false
        )]
        public static SqlDouble Quantile(string Query, double Quantile)
        {
            using (SqlConnection connection = new SqlConnection("context connection=true"))
            {
                try
                {
                    if (Quantile > 1 || Quantile < 0)
                        return SqlDouble.Null;
                    else
                    {
                        string SafeQuery = "select dbo.fMathSafeSqlCommand('" + Query.ToString().Trim() + "'," + MaxNumRows_Stat.ToString().Trim() + ")";
                        SqlCommand cmd = new SqlCommand();
                        cmd.Connection = connection; cmd.CommandText = SafeQuery;
                        connection.Open();
                        SqlDataReader rdr = cmd.ExecuteReader();
                        rdr.Read();
                        string ColName = rdr.GetName(0);
                        SafeQuery = rdr[0].ToString().Trim();
                        if (SafeQuery.StartsWith("error") || ColName == "error_message" || SafeQuery.StartsWith("Syntax is OK"))
                            throw new System.Exception(SafeQuery);
                        if (!SafeQuery.Contains(" as col "))
                        {
                            throw new System.Exception("The query '" + Query + "' should have its returning column aliased with the following exact text: 'as col'. Example: select x as col from Table");
                        }
                        rdr.Close();

                        SafeQuery = "if(" + Quantile.ToString().Trim() + " > 1 or " + Quantile.ToString().Trim() + " <= 0 ) begin SELECT NULL as Variable end else " +

                        "begin declare @counts float;select @counts = a.co from(select count(*) as co from (" + SafeQuery + ") as c where c.col is not NULL) as a;" +
                        "declare @Floor float ; declare @Ceil float;set @Floor = FLOOR(round(" + Quantile.ToString().Trim() + "*@Counts,13));" +
                        "set @Ceil = CEILING(round(" + Quantile.ToString().Trim() + "*@Counts,13)); DECLARE @temp080a table (v float, p float);" +
                        "insert @temp080a  select c.variable as v, c.rownumber/@counts as p from " +
                        "(SELECT q.col as variable,ROW_NUMBER() OVER (ORDER BY q.col) AS RowNumber " +
                        "FROM (" + SafeQuery + ") as q where q.col is not NULL) c where c.rownumber between @FLoor and @CEIL;" +
                        "declare @c bigint ; select @c=a.co from (select count(*) as co from @temp080a) as a; declare @answer float; " +
                        "if(@c<2) begin if(@floor<1) begin set @answer=NULL  end else begin select top 1 @answer = v from @temp080a end end " +
                        "else begin declare @V1 float declare @V2 float declare @P1 float declare @P2 float " +
                        "select top 1 @V1 = v from @temp080a order by p asc " +
                        "select top 1 @V2 = v from @temp080a order by p desc " +
                        "select top 1 @P1 = p from @temp080a order by p asc " +
                        "select top 1 @P2 = p from @temp080a order by p desc " +
                        "set @answer = @v1+ ((@v2-@v1)*(" + Quantile.ToString().Trim() + "-@p1)/(@p2-@p1)) end;" +
                        "if(abs(@answer)<=1e-13) begin select round(@answer,13) as Quantile end else begin select dbo.fMathRound(@answer,13) as Quantile end end";

                        cmd.CommandText = SafeQuery;
                        rdr = cmd.ExecuteReader();
                        rdr.Read();
                        if (rdr[0].ToString() != System.DBNull.Value.ToString())
                            return SqlDouble.Parse(rdr[0].ToString());
                        else
                            return SqlDouble.Null;
                    }
                }
                catch
                {
                    throw;
                }
                finally
                {
                    connection.Close();
                }
            }
        }

        //----------------------------------------------------------------------------------------------------------------------------------------------------

        [SqlFunction(
        Name = "fMathCDF",
        DataAccess = DataAccessKind.Read,
        SystemDataAccess = SystemDataAccessKind.Read,
        IsDeterministic = false,
        IsPrecise = false
        )]
        public static SqlDouble CumulativeDistribution(string Query, double Variable)
        {
            using (SqlConnection connection = new SqlConnection("context connection=true"))
            {
                try
                {
                            
                    string SafeQuery = "select dbo.fMathSafeSqlCommand('" + Query.ToString().Trim() + "'," + MaxNumRows_Stat.ToString().Trim() + ")";
                    SqlCommand cmd = new SqlCommand();
                    cmd.Connection = connection; cmd.CommandText = SafeQuery;
                    connection.Open();
                    SqlDataReader rdr = cmd.ExecuteReader();
                    rdr.Read();
                    string ColName = rdr.GetName(0);
                    SafeQuery = rdr[0].ToString().Trim();
                    if (SafeQuery.StartsWith("error") || ColName == "error_message" || SafeQuery.StartsWith("Syntax is OK"))
                        throw new System.Exception(SafeQuery);
                    if (!SafeQuery.Contains(" as col "))
                    {
                        throw new System.Exception("The query '" + Query + "' should have its returning column aliased with the following exact text: 'as col'. Example: select x as col from Table");
                    }
                    rdr.Close();
                    SafeQuery = "declare @counts float;declare @PartialCounts float;" +
                    "select @counts=a.co, @PartialCounts = a.cop from ( select count(*) as co, sum(case when c.col <=" + Variable.ToString().Trim() + " then 1 else 0 end) as cop from (" + Query + ") as c where c.col is not null) as a;" +
                    "declare @Floor float ; declare @Ceil float;set @Floor = @PartialCounts;" +
                    "set @Ceil = @PartialCounts+1; DECLARE @temp080a table (v float, p float);" +
                    "insert @temp080a  select c.variable as v, c.rownumber/@counts as p from " +
                    "(SELECT q.col as variable, ROW_NUMBER() OVER (ORDER BY q.col) AS RowNumber " +
                    "FROM (" + Query + ") as q where q.col is not null) c where c.rownumber between @FLoor and @CEIL;" +
                    "select @counts=a.co from (select count(*) as co from @temp080a) as a; declare @answer float;" +
                    "if(@counts<2) 	begin if(@Floor<1) begin select @answer= 0 end else begin select top 1 @answer= p from @temp080a end end " +
                    "else begin declare @V1 float declare @V2 float declare @P1 float declare @P2 float " +
                    "select top 1 @V1 = v from @temp080a order by p asc " +
                    "select top 1 @V2 = v from @temp080a order by p desc " +
                    "select top 1 @P1 = p from @temp080a order by p asc " +
                    "select top 1 @P2 = p from @temp080a order by p desc " +
                    "set @answer = @p1+ ((@p2-@p1)*(" + Variable.ToString().Trim() + "-@v1)/(@v2-@v1)) end;" +
                    "select round(@answer,15) as CDF";

                    cmd.CommandText = SafeQuery;
                    rdr = cmd.ExecuteReader();
                    rdr.Read();
                    if (rdr[0].ToString() != System.DBNull.Value.ToString())
                        return SqlDouble.Parse(rdr[0].ToString());
                    else
                        return SqlDouble.Null;
                }
                catch
                {
                    throw;
                }
                finally
                {
                    connection.Close();
                }
            }
        }

        //----------------------------------------------------------------------------------------------------------------------------------------------------

        static public string Revision
        {
            get { return "$Revision: 1..0 $"; }
        }

    }


    [Serializable]
    [SqlUserDefinedAggregate(
    Format.UserDefined,
    IsInvariantToNulls = true,
    IsNullIfEmpty = true,
    IsInvariantToDuplicates = false,
    IsInvariantToOrder = true,
    MaxByteSize = -1)]
    public struct QuantileFunction : IBinarySerialize
    {
        private const int Sig = 13;
        private double Quantile;
        private List<SqlDouble> Values;
        
        public void Init()
        {
            this.Quantile = 0;
            this.Values = new List<SqlDouble>();
        }

        public void Accumulate(SqlDouble value, SqlDouble quantile)
        {
            if (!value.IsNull && quantile <= 1 && quantile >= 0)
            {
                Quantile = (double)quantile;
                Values.Add((double)value);
            }
            else
                return;
        }

        public void Merge(QuantileFunction ListOfValues)
        {
            this.Values.AddRange(ListOfValues.Values);
            this.Quantile = ListOfValues.Quantile;
        }

        public SqlDouble Terminate()
        {
            double c = Values.Count; int i1 = 0;
            SqlDouble Answer = c;
            if (c > 0)
            {
                Mathematics.MathFunctions MATH = new Mathematics.MathFunctions();
                double z = Quantile * c;
                //i1 = (int)MATH.Round(z, 14) - 1;
                i1 = (int)Math.Round(z, 15) - 1;
                if (i1 < 0)
                    Answer = SqlDouble.Null;
                else
                {
                    Values.Sort();
                    if (i1 == Values.Count - 1)
                        Answer = Values[Values.Count - 1];
                    else
                        //Answer = Values[i1] + (Values[i1 + 1] - Values[i1]) * (Quantile - (i1 + 1.0) / c) * c;
                        Answer = Values[i1] + (Values[i1 + 1] - Values[i1]) * (z - (i1 + 1.0));
                    Answer = MATH.Round((double)Answer, Sig);
                }
                return Answer;
            }
            else
                return SqlDouble.Null;
        }

        public void Read(BinaryReader reader)
        {
            this.Quantile = reader.ReadDouble();
            this.Values = new List<SqlDouble>();
            int Length = reader.ReadInt32();
            for (int i = 0; i < Length; i++)
            {
                this.Values.Add(reader.ReadDouble());
            }
        }

        public void Write(BinaryWriter writer)
        {
            writer.Write(Quantile);
            writer.Write(this.Values.Count);
            foreach (double value in this.Values)
            {
                writer.Write(value);
            }
        }

    }


    [Serializable]
    [SqlUserDefinedAggregate(
    Format.UserDefined,
    IsInvariantToNulls = true,
    IsNullIfEmpty = true,
    IsInvariantToDuplicates = false,
    IsInvariantToOrder = true,
    MaxByteSize = -1)]
    public struct CumulativeDistribution : IBinarySerialize
    {
        private const int Sig = 15;
        private double Value;
        private Int32 PartialCounts;
        private List<SqlDouble> Values;

        public void Init()
        {
            this.Value = 0;
            this.PartialCounts = 0;
            this.Values = new List<SqlDouble>();
        }

        public void Accumulate(SqlDouble value, SqlDouble Value_)
        {
            if (!value.IsNull && !Value_.IsNull)
            {
                this.Value = (double)Value_;
                if (value <= Value)
                    PartialCounts++;
                Values.Add(value);

            }
            else
                return;
        }

        public void Merge(CumulativeDistribution ListOfValues)
        {
            this.Value = ListOfValues.Value;
            this.PartialCounts = this.PartialCounts + ListOfValues.PartialCounts;
            this.Values.AddRange(ListOfValues.Values);
        }

        public SqlDouble Terminate()
        {
            if (Values.Count > 0)
            {
                SqlDouble Answer;
                Values.Sort();
                if (Value < Values[0] || PartialCounts == 0)
                    Answer = 0;
                else
                {
                    if (Value >= Values[Values.Count - 1])
                        Answer = 1;
                    else
                    {
                        Mathematics.MathFunctions MATH = new Mathematics.MathFunctions();
                        double p1 = (PartialCounts) / (double)Values.Count;
                        double p2 = (PartialCounts + 1) / (double)Values.Count;
                        Answer = p1 + (p2 - p1) * (Value - Values[PartialCounts - 1]) / (Values[PartialCounts] - Values[PartialCounts - 1]);
                        Answer = (SqlDouble)Math.Round((double)Answer, Sig);
                    }
                }
                return Answer;
            }
            else
                return SqlDouble.Null;
        }

        public void Read(BinaryReader reader)
        {
            Value = reader.ReadDouble();
            PartialCounts = reader.ReadInt32();
            this.Values = new List<SqlDouble>();
            int Length = reader.ReadInt32();
            for (int i = 0; i < Length; i++)
            {
                this.Values.Add(reader.ReadDouble());
            }
        }

        public void Write(BinaryWriter writer)
        {
            writer.Write(Value);
            writer.Write(PartialCounts);
            writer.Write(this.Values.Count);
            foreach (double value in this.Values)
            {
                writer.Write(value);
            }
        }

    
    }









}

